class "Shaper"

require "ShaperWindow"

----------------------------------------------------------------------------------------------------

class "RenoiseScriptingTool" (renoise.Document.DocumentNode)


function RenoiseScriptingTool:__init()

    renoise.Document.DocumentNode.__init(self)
    self:add_property("Name", "Untitled Tool")
    self:add_property("Id", "Unknown Id")

end


----------------------------------------------------------------------------------------------------


function Shaper:__init ()

    local m = RenoiseScriptingTool()
    m:load_from("manifest.xml")
    self.tool_name = m:property("Name").value
    self.tool_id = m:property("Id").value

    self.window = ShaperWindow (self)

end


----------------------------------------------------------------------------------------------------


-- Some terminology:
-- A shape is what the user selected, can contain one or two peaks and is made of up to four parts.
-- A peak is made of either one or two parts.


function Shaper:write_curve (shp_1st, shp_2nd, shp_3rd, shp_4th, alt_peak_index, length, nb_peaks, offset, phase, resolution)

    if renoise.app().window.active_lower_frame ~= renoise.ApplicationWindow.LOWER_FRAME_TRACK_AUTOMATION then
        renoise.app().window.active_lower_frame = renoise.ApplicationWindow.LOWER_FRAME_TRACK_AUTOMATION
    end

    self:remove_points (length, offset)

    -- Scale the phase to the width of a peak
    phase = phase * length / nb_peaks

    -- From a user point of view, a two-part shape count as one, but not for for us
    local parts_per_shape = 1 -- Number of parts
    if shp_2nd then parts_per_shape = parts_per_shape + 1 end
    if shp_3rd then parts_per_shape = parts_per_shape + 1 end
    if shp_4th then parts_per_shape = parts_per_shape + 1 end
    nb_peaks = nb_peaks * parts_per_shape

    local origins = {}
    local peak_scales = {}
    for i = 1, nb_peaks + 1 do
        origins[i] = self:origin_for_step (i, nb_peaks, parts_per_shape)
        peak_scales[i] = self:scale_for_step (i, nb_peaks, parts_per_shape, alt_peak_index, origins[i])
    end
    -- The different parts of a peak need to share the same scale
    for i = 1, nb_peaks do
        local j = (i - 1) % parts_per_shape + 1
        if  j ~= 1 and j ~= alt_peak_index then
            peak_scales[i] = peak_scales[i - 1]
        end
        if  j ~= 1 then
            origins[i] = origins[i - 1]
        end
    end

    -- First pass
    for pos = 0, length - 1/256, resolution do

        -- Calculate the current step, and x coordinate within it
        local x = nb_peaks * (pos + phase) / length
        local st = math.floor (x) + 1
        x = x % 1

        local shape = shp_1st
        if parts_per_shape == 2 and st % 2 == 0 and shp_2nd then
            shape = shp_2nd
        elseif parts_per_shape == 3 and st % 3 == 2 and shp_2nd then
            shape = shp_2nd
        elseif parts_per_shape == 3 and st % 3 == 0 and shp_3rd then
            shape = shp_3rd
        elseif parts_per_shape == 4 and st % 4 == 2 and shp_2nd then
            shape = shp_2nd
        elseif parts_per_shape == 4 and st % 4 == 3 and shp_3rd then
            shape = shp_3rd
        elseif parts_per_shape == 4 and st % 4 == 0 and shp_4th then
            shape = shp_4th
        end

        --~ local shape_step = math.floor ((st - 1) / parts_per_shape) + 1
        local o, scl = origins[st], peak_scales[st]

        self:add_point (offset + pos, o + scl * Shaper.shape_functions[shape] (x))

    end

    -- Second pass to enforce the first and last point of each shape
    for st = 1, nb_peaks + 1 do

        local shape = shp_1st
        if parts_per_shape == 2 and st % 2 == 0 and shp_2nd then
            shape = shp_2nd
        elseif parts_per_shape == 3 and st % 3 == 2 and shp_2nd then
            shape = shp_2nd
        elseif parts_per_shape == 3 and st % 3 == 0 and shp_3rd then
            shape = shp_3rd
        elseif parts_per_shape == 4 and st % 4 == 2 and shp_2nd then
            shape = shp_2nd
        elseif parts_per_shape == 4 and st % 4 == 3 and shp_3rd then
            shape = shp_3rd
        elseif parts_per_shape == 4 and st % 4 == 0 and shp_4th then
            shape = shp_4th
        end

        --~ local shape_step = math.floor ((st - 1) / parts_per_shape) + 1
        local o, scl = origins[st], peak_scales[st]

        local p = offset - phase + (st - 1) * (length / nb_peaks)
        if p >= 0 and p < length then
            self:add_point (p, o + scl * Shaper.shape_functions[shape] (0))
        end
        p = offset - phase + st * (length / nb_peaks) - 1/256
        if p > 0 and p < length then
            self:add_point (p, o + scl * Shaper.shape_functions[shape] (1))
        end

    end

end


----------------------------------------------------------------------------------------------------


function Shaper:origin_for_step (part_step, nb_peaks, parts_per_shape)

    local nb_shapes = math.floor ((nb_peaks + 1) / parts_per_shape)
    local shape_step = math.floor ((part_step - 1) / parts_per_shape) + 1

    local o = 0
    if self.window.vb.views.step_origin.value == 2 then

        o = (shape_step - 1) * (self.window.vb.views.origin_dest.value - self.window.vb.views.origin.value)
        if nb_shapes > 1 then
            o = o / (nb_shapes - 1)
        end
        o = o + self.window.vb.views.origin.value

    elseif self.window.vb.views.step_origin.value == 3 then

        local a = self.window.vb.views.origin.value * 1000
        local b = self.window.vb.views.origin_dest.value * 1000
        if a < b then
            o = math.random (a, b) / 1000
        else
            o = math.random (b, a) / 1000
        end

    else
        o = o + self.window.vb.views.origin.value
    end

    return o

end


----------------------------------------------------------------------------------------------------


function Shaper:scale_for_step (part_step, nb_peaks, parts_per_shape, alt_peak_index, origin)

    local nb_shapes = math.floor ((nb_peaks + 1) / parts_per_shape)
    local shape_step = math.floor ((part_step - 1) / parts_per_shape) + 1

    local scl =  0


    if self.window.vb.views.step_peak.value == 2 then

         scl = (shape_step - 1) * (self.window.vb.views.peak_dest.value - self.window.vb.views.peak.value)
        if nb_shapes > 1 then
            scl = scl / (nb_shapes - 1)
        end
        scl = scl + self.window.vb.views.peak.value

    elseif self.window.vb.views.step_peak.value == 3 then

        local a = self.window.vb.views.peak.value * 1000
        local b = self.window.vb.views.peak_dest.value * 1000
        if a < b then
            scl = math.random (a, b) / 1000
        else
            scl = math.random (b, a) / 1000
        end

    else
        scl = scl + self.window.vb.views.peak.value
    end


    if
        self.window.vb.views.alternate_peaks.value
        and (part_step - 1) % parts_per_shape >= alt_peak_index - 1
    then

        scl = 0
        if self.window.vb.views.step_alt_peak.value == 2 then

            scl = (shape_step - 1) * (self.window.vb.views.alt_peak_dest.value - self.window.vb.views.alt_peak.value)
            if nb_shapes > 1 then
                scl = scl / (nb_shapes - 1)
            end
            scl = scl + self.window.vb.views.alt_peak.value

        elseif self.window.vb.views.step_alt_peak.value == 3 then

            local a = self.window.vb.views.alt_peak.value * 1000
            local b = self.window.vb.views.alt_peak_dest.value * 1000
            if a < b then
                scl = math.random (a, b) / 1000
            else
                scl = math.random (b, a) / 1000
            end

        else
            scl = scl + self.window.vb.views.alt_peak.value
        end

    end

    scl = scl - origin

    return scl

end


----------------------------------------------------------------------------------------------------


function Shaper:remove_points (s, off)

    local r = renoise.song().selected_parameter
    if not r then
        renoise.app():show_status ("Shaper Error: No parameter selected")
        return
    end

    local st = renoise.song().selected_track_index
    local ss = renoise.song().selected_sequence_index
    local sp = renoise.song().selected_pattern_index

    local sl = renoise.song().selected_line_index
    local nl = renoise.song().patterns[sp].number_of_lines

    local a = renoise.song().patterns[sp].tracks[st]:find_automation(r)

    local dl = off
    for i = 0, s - 1/256, 1/256 do

        dl = dl + 1/256
        while sl + dl > nl + 1 do
            ss = ss + 1
            sp = renoise.song().sequencer.pattern_sequence[ss]
            if not sp then
                return
            end
            nl = renoise.song().patterns[sp].number_of_lines
            a = renoise.song().patterns[sp].tracks[st]:find_automation(r)
            sl = 1
            dl = 0
        end

        if a then
            if a:has_point_at (sl + dl) then
                a:remove_point_at (sl + dl)
            end
        end

    end


    -- Alternate method (same speed)

    --~ while s > 0 do

        --~ if a then
            --~ local points = a.points
            --~ local new_points = {}
            --~ for i, point in pairs(points) do
                --~ if point.time < sl + dl or point.time > sl + dl + s - 1/256 then
                    --~ new_points[#new_points + 1] = point
                --~ end
            --~ end
            --~ a.points = new_points
        --~ end

        --~ s = s - (nl - sl)

        --~ if s > 0 then
            --~ ss = ss + 1
            --~ sp = renoise.song().sequencer.pattern_sequence[ss]
            --~ if not sp then
                --~ return
            --~ end
            --~ nl = renoise.song().patterns[sp].number_of_lines
            --~ a = renoise.song().patterns[sp].tracks[st]:find_automation(r)
            --~ sl = 1
            --~ dl = 0
        --~ end

    --~ end

end


----------------------------------------------------------------------------------------------------


local function clamp (v)
    if v < 0 then v = 0
    elseif v > 1 then v = 1
    end
    return v
end


function Shaper:add_point (dl, v)

    local r = renoise.song().selected_parameter
    if not r then
        renoise.app():show_status ("Shaper Error: No parameter selected")
        return
    end

    local st = renoise.song().selected_track_index
    local ss = renoise.song().selected_sequence_index
    local sp = renoise.song().selected_pattern_index

    local sl = renoise.song().selected_line_index
    local nl = renoise.song().patterns[sp].number_of_lines
    while sl + dl > nl + 1 do
        ss = ss + 1
        sp = renoise.song().sequencer.pattern_sequence[ss]
        if not sp then
            return
        end
        dl = dl - (nl - sl + 1)
        nl = renoise.song().patterns[sp].number_of_lines
        sl = 1
    end

    local a = renoise.song().patterns[sp].tracks[st]:find_automation(r)
    if not a then
        a = renoise.song().patterns[sp].tracks[st]:create_automation (r)
    end

    a:add_point_at (sl + dl, clamp(v))

end


----------------------------------------------------------------------------------------------------


-- All shape functions map from [0, 1] to [0, 1]
Shaper.shape_functions =
{
    square_down = function (x) return 0 end,
    square_up = function (x) return 1 end,

    linear_up =  function (x) return x end,
    linear_down = function (x) return 1 - x end,

    step_up = function (x)
        local st = math.floor (x / 0.5)
        return st * 0.5
    end,
    step_down = function (x)
        local st = math.floor ((1 - x) / 0.5)
        if x < 0.996 then
            return st * 0.5 + 0.5
        else
            return 0
        end
    end,

    wave_up = function (x) return (1 - math.cos (x * math.pi)) / 2 end,
    wave_down = function (x) return math.cos (x * math.pi) / 2 + 0.5 end,

    concave_up = function (x) return 1 - math.sin (math.acos (x)) end,
    concave_down = function (x) return 1 - math.sin (math.acos (1 - x)) end,

    convex_up = function (x) return math.sin (math.acos (1 - x)) end,
    convex_down = function (x) return math.sin (math.acos (x)) end,

    halfsine_up = function (x) return 0.5 + math.sin (x * math.pi) / 2 end,
    halfsine_down = function (x) return 0.5 - math.sin(x * math.pi) / 2 end,

    convex_sin_up = function (x) return math.sin(x * math.pi / 2) end,
    convex_sin_down = function (x) return math.cos(x * math.pi / 2) end,

    concave_sin_up = function (x) return 1 - math.cos(x * math.pi / 2) end,
    concave_sin_down = function (x) return 1 - math.sin(x * math.pi / 2) end,
}


----------------------------------------------------------------------------------------------------
