class "ShaperWindow"


----------------------------------------------------------------------------------------------------


function ShaperWindow:__init (shaper)

    self.shaper = shaper
    self.vb = renoise.ViewBuilder ()

    self.shapes = { "1_off", "2_off", "3_off", "4_off" }

    renoise.tool():add_menu_entry {
        name = "Main Menu:Tools:Automasher...",
        invoke = function () self:show_dialog () end
    }

    renoise.tool():add_menu_entry {
        name = "Track Automation:Automasher...",
        invoke = function () self:show_dialog () end
    }

    renoise.tool():add_keybinding {
        name = "Global:Tools:Automasher...",
        invoke = function () self:show_dialog () end
    }

end


----------------------------------------------------------------------------------------------------


function ShaperWindow:show_dialog ()

    if self.dialog and self.dialog.visible then
        self.dialog:show ()
        return
    end

    if not self.dialog_content then
        self.dialog_content = self:gui ()
    end

    local kh = function (d, k) return self:key_handler (d, k) end
    self.dialog = renoise.app():show_custom_dialog ("Automation Shaper", self.dialog_content, kh)

end


----------------------------------------------------------------------------------------------------


function ShaperWindow:key_handler (dialog, key)

    local p = renoise.song().selected_pattern_index
    local t = renoise.song().selected_track_index
    local l = renoise.song().selected_line_index
    local nl = renoise.song().patterns[p].number_of_lines

    -- close on escape...
    if key.modifiers == "" and key.name == "esc" then
        dialog:close()


    elseif key.name == "up" then
        local s = renoise.song().selected_sequence_index
        if l > 1 then
            l = l - 1
        elseif s > 1 then
            s = s - 1
            renoise.song().selected_sequence_index = s
            l = renoise.song().selected_pattern.number_of_lines
        end
        renoise.song().selected_line_index = l

    elseif key.name == "down" then
        if l + 1 <= nl then
            l = l + 1
        else
            local s = renoise.song().selected_sequence_index
            s = s + 1
            if renoise.song().sequencer.pattern_sequence[s] then
                renoise.song().selected_sequence_index = s
                l = 1
            end
        end
        renoise.song().selected_line_index = l

    elseif key.name == "space" then
        if renoise.song().transport.playing then
            renoise.song().transport:stop()
        else
            renoise.song().transport:start_at(1)
        end

    else
        return key

--~
    --~ elseif key.modifiers == "control" and key.name == "z" then
        --~ renoise.song():undo ()
    --~ elseif key.modifiers == "shift + control" and key.name == "z" then
        --~ renoise.song():redo ()
    --~ elseif key.modifiers == "control" and key.name == "y" then
        --~ renoise.song():redo ()

    end

end


----------------------------------------------------------------------------------------------------


function ShaperWindow:on_shape_selected (part, id)

        self.vb.views[self.shapes[part]].bitmap = "Icons/" .. string.sub (self.shapes[part], 3) .. ".png"

        self.vb.views[id].bitmap = "Icons/" .. string.sub (id, 3) .. "_selected.png"
        self.shapes[part] = id

end


----------------------------------------------------------------------------------------------------


function ShaperWindow:auto_update ()

    if self.vb.views.auto_update.value then
        self:on_write_pressed ()
    end

end


----------------------------------------------------------------------------------------------------


function ShaperWindow:on_write_pressed ()

    -- How many times the shape should repeat inside the slice
    local len = self.vb.views.length.value
    local nbs = self.vb.views.number_of_shapes.value
    local off = 0 -- self.vb.views.offset.value
    local phs = self.vb.views.phase.value
    local res = 1 / self.vb.views.resolution.value

    local shp = {}
    local api = nil -- Alternate peak first index
    if self.shapes[1] ~= "1_off" then
        shp[#shp + 1] = string.sub (self.shapes[1], 3)
    end
    if self.shapes[2] ~= "2_off" then
        shp[#shp + 1] = string.sub (self.shapes[2], 3)
    end
    if self.vb.views.alternate_peaks.value and self.shapes[3] ~= "3_off" then
        shp[#shp + 1] = string.sub (self.shapes[3], 3)
        api = #shp
    end
    if self.vb.views.alternate_peaks.value and self.shapes[4] ~= "4_off" then
        shp[#shp + 1] = string.sub (self.shapes[4], 3)
        if not api then
            api = #shp
        end
    end

    if not shp[1] then
        return
    end

    if not api then
        api = 99 ---TODO
    end

    self.shaper:write_curve (shp[1], shp[2], shp[3], shp[4], api, len, nbs, off, phs, res)

end


----------------------------------------------------------------------------------------------------


function ShaperWindow:percent_value_tonumber (v)

    local number_part = string.match (v, "%d*%.?%,?%d*")

    local v2 = tonumber(number_part)
    if not v2 then
        return
    end
    v2 = v2 / 100

    if v2 < 0 then
        v2 = 0
    elseif v2 > 1 then
        v2 = 1
    end

    return v2

end


function ShaperWindow:percent_value_tostring (v)

    return string.format("%.3f %%", v * 100)

end


----------------------------------------------------------------------------------------------------


function ShaperWindow:gui ()

    local vb = self.vb

    local result = vb:column
    {
        style = "body",
        margin = 8,
        spacing = 8,
        uniform = true,

        vb:column
        {
            style = "group",
            margin = 4,
            vb:row
            {
                vb:text { text = "Length", width = 70 },
                vb:valuebox { id = "length", value = 2, min = 1, max = 9999, notifier = function () self:auto_update () end, tooltip = "Total length of the curve, in lines" },
                vb:row { width = 45 },
                vb:text { text = "Shapes", width=45 },
                vb:valuebox { id = "number_of_shapes", value = 1, min = 1, max = 9999, notifier = function () self:auto_update () end, tooltip = "Number of shapes in the curve" },
            },

            -- The "Offset" feature is disabled

            --~ vb:row
            --~ {
                --~ vb:text { text = "Offset", width=70 },
                --~ vb:slider
                --~ {
                    --~ id = "offset_slider",
                    --~ value = 0, min = 0, max = 1,
                    --~ notifier = function (v) self.vb.views.offset.value = v end,
                    --~ width=150,
                --~ },
                --~ vb:valuefield
                --~ {
                    --~ id = "offset",
                    --~ value = 0, min = 0, max = 1,
                    --~ tonumber = function (v) return self:percent_value_tonumber (v) end,
                    --~ tostring = function (v) return self:percent_value_tostring (v) end,
                    --~ notifier = function (v) self.vb.views.offset_slider.value = v end,
                --~ },
                --~ tooltip = "Offset the whole curve by a fraction of a line",
            --~ },

            vb:row
            {
                vb:text { text = "Phase", width=70 },
                vb:slider
                {
                    id = "phase_slider",
                    value = 0, min = 0, max = 1,
                    notifier = function (v) self.vb.views.phase.value = v ; self:auto_update() end,
                    width=150,
                },
                vb:valuefield
                {
                    id = "phase",
                    value = 0, min = 0, max = 1,
                    tonumber = function (v) return self:percent_value_tonumber (v) end,
                    tostring = function (v) return self:percent_value_tostring (v) end,
                    notifier = function (v) self.vb.views.phase_slider.value = v ; self:auto_update() end,
                },
            },


            vb:row
            {
                vb:text { text = "Points / Line", width = 70 },
                vb:valuebox { id = "resolution", value = 8, min = 1, max = 256, notifier = function () self:auto_update () end, tooltip = "Number of automation points per line" },

                --~ -- The "auto-update" feature is disabled
                vb:row { width = 45 },

                vb:checkbox { id = "auto_update", visible = true },
                vb:text { text = "Auto-update" },
            },
        },

        vb:column
        {
            style = "group",
            margin = 4,
            vb:row
            {
                vb:text { text = "Origin", width=70 },
                vb:slider
                {
                    id = "origin_slider",
                    value = 0, min = 0, max = 1,
                    notifier = function (v) self.vb.views.origin.value = v ; self:auto_update() end,
                    width=150,
                },
                vb:valuefield
                {
                    id = "origin",
                    value = 0, min = 0, max = 1,
                    tonumber = function (v) return self:percent_value_tonumber (v) end,
                    tostring = function (v) return self:percent_value_tostring (v) end,
                    notifier = function (v) self.vb.views.origin_slider.value = v ; self:auto_update() end,
                },
            },

            vb:row
            {
                --~ vb:horizontal_aligner
                --~ {
                    --~ vb:checkbox
                    --~ {
                        --~ id = "step_origin",
                        --~ notifier = function ()
                            --~ local a = self.vb.views.step_origin.value
                            --~ self.vb.views.origin_dest_slider.visible = a
                            --~ self.vb.views.origin_dest.visible = a
                            --~ self:auto_update()
                        --~ end
                    --~ },
                    --~ vb:text { text = "Step to", },
                    --~ width = 70,
                --~ },
                vb:popup
                {
                    id = "step_origin",
                    items = { "Fixed", "Step to", "Random" },
                    notifier = function ()
                        local a = self.vb.views.step_origin.value > 1
                        self.vb.views.origin_dest_slider.visible = a
                        self.vb.views.origin_dest.visible = a
                        self:auto_update()
                    end,
                    width = 70,
                },
                vb:slider
                {
                    id = "origin_dest_slider",
                    value = 0, min = 0, max = 1,
                    notifier = function (v) self.vb.views.origin_dest.value = v ; self:auto_update() end,
                    width = 150,
                    visible = false,
                },
                vb:valuefield
                {
                    id = "origin_dest",
                    value = 0, min = 0, max = 1,
                    tonumber = function (v) return self:percent_value_tonumber (v) end,
                    tostring = function (v) return self:percent_value_tostring (v) end,
                    notifier = function (v) self.vb.views.origin_dest_slider.value = v ; self:auto_update() end,
                    visible = false,
                },
            },

        },

        vb:column
        {
            style = "group",
            margin = 4,

            vb:row
            {
                vb:text { text = "Peak", width=70 },
                vb:slider
                {
                    id = "peak_slider",
                    value = 1, min = 0, max = 1,
                    notifier = function (v) self.vb.views.peak.value = v ; self:auto_update() end,
                    width=150,
                },
                vb:valuefield
                {
                    id = "peak",
                    value = 1, min = 0, max = 1,
                    tonumber = function (v) return self:percent_value_tonumber (v) end,
                    tostring = function (v) return self:percent_value_tostring (v) end,
                    notifier = function (v) self.vb.views.peak_slider.value = v ; self:auto_update() end,
                },
            },

            vb:row
            {
                --~ vb:horizontal_aligner
                --~ {
                    --~ vb:checkbox
                    --~ {
                        --~ id = "step_peak",
                        --~ notifier = function ()
                            --~ local a = self.vb.views.step_peak.value
                            --~ self.vb.views.peak_dest_slider.visible = a
                            --~ self.vb.views.peak_dest.visible = a
                            --~ self:auto_update()
                        --~ end
                    --~ },
                    --~ vb:text { text = "Step to", },
                    --~ width = 70,
                --~ },
                vb:popup
                {
                    id = "step_peak",
                    items = { "Fixed", "Step to", "Random" },
                    notifier = function ()
                        local a = self.vb.views.step_peak.value > 1
                        self.vb.views.peak_dest_slider.visible = a
                        self.vb.views.peak_dest.visible = a
                        self:auto_update()
                    end,
                    width = 70,
                },
                vb:slider
                {
                    id = "peak_dest_slider",
                    value = 1, min = 0, max = 1,
                    notifier = function (v) self.vb.views.peak_dest.value = v ; self:auto_update() end,
                    width = 150,
                    visible = false,
                },
                vb:valuefield
                {
                    id = "peak_dest",
                    value = 1, min = 0, max = 1,
                    tonumber = function (v) return self:percent_value_tonumber (v) end,
                    tostring = function (v) return self:percent_value_tostring (v) end,
                    notifier = function (v) self.vb.views.peak_dest_slider.value = v ; self:auto_update() end,
                    visible = false,
                },
            },

            vb:row { height = 8 },


            vb:horizontal_aligner
            {
                spacing = 16,
                mode = "distribute",

                vb:column
                {
                    margin = 4,
                    spacing = 4,

                    vb:row
                    {
                        spacing = 4,
                        vb:bitmap { id = "1_off", bitmap = "Icons/off_selected.png", notifier = function () self:on_shape_selected (1, "1_off") ; self:auto_update() end, },
                        vb:bitmap { id = "1_square_down", bitmap = "Icons/square_down.png", notifier = function () self:on_shape_selected (1, "1_square_down") ; self:auto_update() end, },
                        vb:bitmap { id = "1_step_up", bitmap = "Icons/step_up.png", notifier = function () self:on_shape_selected (1, "1_step_up") ; self:auto_update() end, },
                    },

                    vb:row
                    {
                        spacing = 4,
                        vb:bitmap { id = "1_convex_sin_up", bitmap = "Icons/convex_sin_up.png", notifier = function () self:on_shape_selected (1, "1_convex_sin_up") ; self:auto_update() end, },
                        vb:bitmap { id = "1_wave_up", bitmap = "Icons/wave_up.png", notifier = function () self:on_shape_selected (1, "1_wave_up") ; self:auto_update() end, },
                        vb:bitmap { id = "1_concave_sin_up", bitmap = "Icons/concave_sin_up.png", notifier = function () self:on_shape_selected (1, "1_concave_sin_up") ; self:auto_update() end, },
                    },

                    vb:row
                    {
                        spacing = 4,
                        vb:bitmap { id = "1_convex_up", bitmap = "Icons/convex_up.png", notifier = function () self:on_shape_selected (1, "1_convex_up") ; self:auto_update() end, },
                        vb:bitmap { id = "1_linear_up", bitmap = "Icons/linear_up.png", notifier = function () self:on_shape_selected (1, "1_linear_up") ; self:auto_update() end, },
                        vb:bitmap { id = "1_concave_up", bitmap = "Icons/concave_up.png", notifier = function () self:on_shape_selected (1, "1_concave_up") ; self:auto_update() end, },
                    },

                },


                vb:column
                {
                    margin = 4,
                    spacing = 4,

                    vb:row
                    {
                        spacing = 4,
                        vb:bitmap { id = "2_off", bitmap = "Icons/off_selected.png", notifier = function () self:on_shape_selected (2, "2_off") ; self:auto_update() end, },
                        vb:bitmap { id = "2_square_up", bitmap = "Icons/square_up.png", notifier = function () self:on_shape_selected (2, "2_square_up") ; self:auto_update() end, },
                        vb:bitmap { id = "2_step_down", bitmap = "Icons/step_down.png", notifier = function () self:on_shape_selected (2, "2_step_down") ; self:auto_update() end, },
                    },

                    vb:row
                    {
                        spacing = 4,
                        vb:bitmap { id = "2_convex_sin_down", bitmap = "Icons/convex_sin_down.png", notifier = function () self:on_shape_selected (2, "2_convex_sin_down") ; self:auto_update() end, },
                        vb:bitmap { id = "2_wave_down", bitmap = "Icons/wave_down.png", notifier = function () self:on_shape_selected (2, "2_wave_down") ; self:auto_update() end, },
                        vb:bitmap { id = "2_concave_sin_down", bitmap = "Icons/concave_sin_down.png", notifier = function () self:on_shape_selected (2, "2_concave_sin_down") ; self:auto_update() end, },
                    },

                    vb:row
                    {
                        spacing = 4,
                        vb:bitmap { id = "2_convex_down", bitmap = "Icons/convex_down.png", notifier = function () self:on_shape_selected (2, "2_convex_down") ; self:auto_update() end, },
                        vb:bitmap { id = "2_linear_down", bitmap = "Icons/linear_down.png", notifier = function () self:on_shape_selected (2, "2_linear_down") ; self:auto_update() end, },
                        vb:bitmap { id = "2_concave_down", bitmap = "Icons/concave_down.png", notifier = function () self:on_shape_selected (2, "2_concave_down") ; self:auto_update() end, },
                    },

                },

            },

        },



        -- Alternate Peaks -------------------------------------------------------------------------
        vb:row
        {
            vb:checkbox
            {
                id = "alternate_peaks",
                notifier = function ()
                    local a = self.vb.views.alternate_peaks.value
                    self.vb.views.alt_peak_group.visible = a
                    self:auto_update()
                end
            },
            vb:text { text = "Alternate peaks", },
        },


        vb:column
        {
            id = "alt_peak_group",
            visible = false,
            style = "group",
            margin = 4,

            vb:row
            {
                vb:text { text = "Alt. Peak", width = 70 },
                vb:slider
                {
                    id = "alt_peak_slider",
                    value = 1, min = 0, max = 1,
                    notifier = function (v) self.vb.views.alt_peak.value = v ; self:auto_update() end,
                    width=150,
                },
                vb:valuefield
                {
                    id = "alt_peak",
                    value = 1, min = 0, max = 1,
                    tonumber = function (v) return self:percent_value_tonumber (v) end,
                    tostring = function (v) return self:percent_value_tostring (v) end,
                    notifier = function (v) self.vb.views.alt_peak_slider.value = v ; self:auto_update() end,
                },
            },

            vb:row
            {
                --~ vb:horizontal_aligner
                --~ {
                    --~ vb:checkbox
                    --~ {
                        --~ id = "step_alt_peak",
                        --~ notifier = function ()
                            --~ local a = self.vb.views.step_alt_peak.value
                            --~ self.vb.views.alt_peak_dest_slider.visible = a
                            --~ self.vb.views.alt_peak_dest.visible = a
                            --~ self:auto_update()
                        --~ end,
                    --~ },
                    --~ vb:text { text = "Step to", },
                    --~ width = 70,
                --~ },
                vb:popup
                {
                    id = "step_alt_peak",
                    items = { "Fixed", "Step to", "Random" },
                    notifier = function ()
                        local a = self.vb.views.step_alt_peak.value > 1
                        self.vb.views.alt_peak_dest_slider.visible = a
                        self.vb.views.alt_peak_dest.visible = a
                        self:auto_update()
                    end,
                    width = 70,
                },
                vb:slider
                {
                    id = "alt_peak_dest_slider",
                    value = 1, min = 0, max = 1,
                    notifier = function (v) self.vb.views.alt_peak_dest.value = v ; self:auto_update() end,
                    width=150,
                    visible = false,
                },
                vb:valuefield
                {
                    id = "alt_peak_dest",
                    value = 1, min = 0, max = 1,
                    tonumber = function (v) return self:percent_value_tonumber (v) end,
                    tostring = function (v) return self:percent_value_tostring (v) end,
                    notifier = function (v) self.vb.views.alt_peak_dest_slider.value = v ; self:auto_update() end,
                    visible = false,
                },
            },


            vb:row { height = 16, },


            vb:horizontal_aligner
            {
                spacing = 16,
                mode = "distribute",

                vb:column
                {
                    margin = 4,
                    spacing = 4,

                    vb:row
                    {
                        spacing = 4,
                        vb:bitmap { id = "3_off", bitmap = "Icons/off_selected.png", notifier = function () self:on_shape_selected (3, "3_off") ; self:auto_update() end, },
                        vb:bitmap { id = "3_square_down", bitmap = "Icons/square_down.png", notifier = function () self:on_shape_selected (3, "3_square_down") ; self:auto_update() end, },
                        vb:bitmap { id = "3_step_up", bitmap = "Icons/step_up.png", notifier = function () self:on_shape_selected (3, "3_step_up") ; self:auto_update() end, },
                    },

                    vb:row
                    {
                        spacing = 4,
                        vb:bitmap { id = "3_convex_sin_up", bitmap = "Icons/convex_sin_up.png", notifier = function () self:on_shape_selected (3, "3_convex_sin_up") ; self:auto_update() end, },
                        vb:bitmap { id = "3_wave_up", bitmap = "Icons/wave_up.png", notifier = function () self:on_shape_selected (3, "3_wave_up") ; self:auto_update() end, },
                        vb:bitmap { id = "3_concave_sin_up", bitmap = "Icons/concave_sin_up.png", notifier = function () self:on_shape_selected (3, "3_concave_sin_up") ; self:auto_update() end, },
                    },

                    vb:row
                    {
                        spacing = 4,
                        vb:bitmap { id = "3_convex_up", bitmap = "Icons/convex_up.png", notifier = function () self:on_shape_selected (3, "3_convex_up") ; self:auto_update() end, },
                        vb:bitmap { id = "3_linear_up", bitmap = "Icons/linear_up.png", notifier = function () self:on_shape_selected (3, "3_linear_up") ; self:auto_update() end, },
                        vb:bitmap { id = "3_concave_up", bitmap = "Icons/concave_up.png", notifier = function () self:on_shape_selected (3, "3_concave_up") ; self:auto_update() end, },
                    },

                },


                vb:column
                {
                    margin = 4,
                    spacing = 4,

                    vb:row
                    {
                        spacing = 4,
                        vb:bitmap { id = "4_off", bitmap = "Icons/off_selected.png", notifier = function () self:on_shape_selected (4, "4_off") ; self:auto_update() end, },
                        vb:bitmap { id = "4_square_up", bitmap = "Icons/square_up.png", notifier = function () self:on_shape_selected (4, "4_square_up") ; self:auto_update() end, },
                        vb:bitmap { id = "4_step_down", bitmap = "Icons/step_down.png", notifier = function () self:on_shape_selected (4, "4_step_down") ; self:auto_update() end, },
                    },

                    vb:row
                    {
                        spacing = 4,
                        vb:bitmap { id = "4_convex_sin_down", bitmap = "Icons/convex_sin_down.png", notifier = function () self:on_shape_selected (4, "4_convex_sin_down") ; self:auto_update() end, },
                        vb:bitmap { id = "4_wave_down", bitmap = "Icons/wave_down.png", notifier = function () self:on_shape_selected (4, "4_wave_down") ; self:auto_update() end, },
                        vb:bitmap { id = "4_concave_sin_down", bitmap = "Icons/concave_sin_down.png", notifier = function () self:on_shape_selected (4, "4_concave_sin_down") ; self:auto_update() end, },
                    },

                    vb:row
                    {
                        spacing = 4,
                        vb:bitmap { id = "4_convex_down", bitmap = "Icons/convex_down.png", notifier = function () self:on_shape_selected (4, "4_convex_down") ; self:auto_update() end, },
                        vb:bitmap { id = "4_linear_down", bitmap = "Icons/linear_down.png", notifier = function () self:on_shape_selected (4, "4_linear_down") ; self:auto_update() end, },
                        vb:bitmap { id = "4_concave_down", bitmap = "Icons/concave_down.png", notifier = function () self:on_shape_selected (4, "4_concave_down") ; self:auto_update() end, },
                    },

                },


            },

        },


        vb:horizontal_aligner
        {
            mode = "center",
            vb:button { id = "write", notifier = function () self:on_write_pressed () end, text = "Write Curve", width = 200, height = 32 },
        },

        --~ vb:column
        --~ {
            --~ style = "group",
            --~ margin = 4,
            --~ vb:row
            --~ {
                --~ vb:text { text = "Modify: ", width=70 },
                --~ vb:popup { items = { "Set", "Add", "Sub", "Mul", "Div", "Humanize" }, value = 4, width = 50 },
                --~ vb:valuefield {},
                --~ vb:button { text = "Apply" }
            --~ },
        --~ },

    }

    return result
end
